/*
 * Copyright (C) 2004 by r0lZ.
 *
 * This program is part of PgcEdit.
 *
 * PgcEdit is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#define PROGRAM_VERSION "0.6"
#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#if defined(HAVE_INTTYPES_H)
#include <inttypes.h>
#elif defined(HAVE_STDINT_H)
#include <stdint.h>
#endif

#include <dvdread/dvd_reader.h>
#include <dvdread/ifo_types.h>
#include <dvdread/ifo_read.h>
#include <dvdread/dvd_udf.h>
#include <dvdread/nav_read.h>
/*
#include <dvdread/nav_print.h>
#include <dvdread/ifo_print.h>
*/

/**
 * Returns true if the pack is a NAV pack.  This check is clearly insufficient,
 * and sometimes we incorrectly think that valid other packs are NAV packs.  I
 * need to make this stronger.
 */
int is_nav_pack( unsigned char *buffer )
{
    return ( buffer[ 41 ] == 0xbf && buffer[ 1027 ] == 0xbf );
}

static void tcl_navPrint_BTNIT(hl_gi_t *hl_gi_table ,btni_t *btni_table, int vobid, int cellid, int vtsnum) {
  int i, j, b, btn_ns, numgroups;
  char head[] = "set ::buttons::butinfo(";
  vm_cmd_t *cmd;

  btn_ns = hl_gi_table->btn_ns;
  numgroups = hl_gi_table->btngr_ns;
  printf( "%s%d,%d,%d,numbuts) %d\n",         head, vtsnum, vobid, cellid, btn_ns);
  if (btn_ns == 0) {
  	printf( "\n" );
  	return;
  }
  printf( "%s%d,%d,%d,numgroups) %d\n",       head, vtsnum, vobid, cellid, numgroups);
  printf( "%s%d,%d,%d,startbutn) %d\n",       head, vtsnum, vobid, cellid, hl_gi_table->btn_ofn);
  printf( "%s%d,%d,%d,forceselectbutn) %d\n", head, vtsnum, vobid, cellid, hl_gi_table->fosl_btnn);
  printf( "%s%d,%d,%d,forceactionbutn) %d\n", head, vtsnum, vobid, cellid, hl_gi_table->foac_btnn);
  printf( "\n");

  for (i = 1; i <= numgroups; i++) {
  	switch (i) {
  	  case 1:
		printf( "%s%d,%d,%d,1,grouptype) %d\n\n", head, vtsnum, vobid, cellid, hl_gi_table->btngr1_dsp_ty);
		break;
  	  case 2:
		printf( "%s%d,%d,%d,2,grouptype) %d\n\n", head, vtsnum, vobid, cellid, hl_gi_table->btngr2_dsp_ty);
		break;
  	  case 3:
		printf( "%s%d,%d,%d,3,grouptype) %d\n\n", head, vtsnum, vobid, cellid, hl_gi_table->btngr3_dsp_ty);
		break;
  	}
    for (j = 1; j <= (36 / numgroups); j++) {
      if (j <= btn_ns) {
        btni_t *btni = &btni_table[(36 / numgroups) * (i-1) + (j-1)];
	    printf( "%s%d,%d,%d,%d,%d,autoaction) %d\n", head, vtsnum, vobid, cellid, i, j, btni->auto_action_mode);
	    printf( "%s%d,%d,%d,%d,%d,coords) {%d %d - %d %d}\n", head, vtsnum, vobid, cellid, i, j, btni->x_start, btni->y_start, btni->x_end, btni->y_end);
	    printf( "%s%d,%d,%d,%d,%d,udlr) {%d %d %d %d}\n", head, vtsnum, vobid, cellid, i, j, btni->up, btni->down, btni->left, btni->right);
	    printf( "%s%d,%d,%d,%d,%d,command) {", head, vtsnum, vobid, cellid, i, j);
	    cmd = &btni->cmd;
	    for (b = 0; b < 8; b++) {
	      printf( "%02X", cmd->bytes[b]);
	      if ( b < 7 ) {
	      	printf( " ");
	      } else {
	      	printf( "}\n\n");
	      }
	    }
	  }
	}
  }
}



int main( int argc, char **argv )
{
	pci_t pci_pack;
    int len, cur_cell;
    unsigned int cur_pack;
    unsigned char *data;
    dvd_reader_t *dvd;
    dvd_file_t *vts;
    ifo_handle_t *vts_file;
    int vtsnum;
    int ok, buttons_found;
    int entries;
    int last_vob_id, last_cell_id;
    char head[] = "set ::buttons::butinfo(";


    /**
     * Usage.
     */
    if ( argc != 3 ) {
        fprintf( stderr, "Tcl Button Dump version " PROGRAM_VERSION " - (c) 2004 by r0lZ.\n"
        				 "Usage: %s <dvd path> <vts number>\n"
                         "Where <vts number> may be 0 for VMGM dump.\n",
			             argv[ 0 ] );
        return -1;
    }

	vtsnum = atoi( argv[ 2 ] );
	data = ( unsigned char * ) malloc( 1024 * DVD_VIDEO_LB_LEN );

    /**
     * Open the disc.
     */
    dvd = DVDOpen( argv[ 1 ] );
    if ( !dvd ) {
  		printf( "%s%d,errorcode) 2\n", head, vtsnum );
  		printf( "%s%d,errormsg) {Couldn't open DVD: %s}\n", head, vtsnum, argv[ 1 ] );
        return 0;
    }


    /**
     * Load the VTS information for the vts set.
     */
    vts_file = ifoOpen( dvd, vtsnum );

    if ( !vts_file ) {
  		printf( "%s%d,errorcode) 3\n", head, vtsnum );
  		printf( "%s%d,errormsg) {Can't open the VTS %d IFO file.}\n", head, vtsnum, vtsnum );
        DVDClose( dvd );
        return 0;
    }


    /**
     * We've got enough info, time to open the title set data.
     */
    vts = DVDOpenFile( dvd, vtsnum, DVD_READ_MENU_VOBS );
    if ( !vts ) {
  		printf( "%s%d,errorcode) 4\n", head, vtsnum );
  		printf( "%s%d,errormsg) {Can't open VTS %d menu VOB.}\n", head, vtsnum, vtsnum );
        ifoClose( vts_file );
        DVDClose( dvd );
        return 0;
    }

	ok = ifoRead_C_ADT( vts_file );
	if ( ! ok ) {
  		printf( "%s%d,errorcode) 5\n", head, vtsnum );
  		printf( "%s%d,errormsg) {Can't open VTS %d C_ADT table.}\n", head, vtsnum, vtsnum );
        ifoClose( vts_file );
        DVDClose( dvd );
        return 0;
	}

	printf( "# PgcEdit menu buttons dump (temp file)\n" );
	printf( "# button_dump v" PROGRAM_VERSION "\n\n" );

    /**
     * Browse all VOB/Cell ID in this VOB.
     */
    last_vob_id = 1;
    last_cell_id = 1;

    entries = (vts_file->menu_c_adt->last_byte + 1 - C_ADT_SIZE)/12;

    printf( "# vts_file->menu_c_adt->nr_of_vobs = %d\n", vts_file->menu_c_adt->nr_of_vobs );
    printf( "# vts_file->menu_c_adt->last_byte  = %d\n", vts_file->menu_c_adt->last_byte );
    printf( "# -> Number of entries  = %d\n", entries );
	/*
	for(cur_cell = 0; cur_cell < entries; cur_cell++) {
		printf("# VOB ID: %3i, Cell ID: %3i, ",
				vts_file->menu_c_adt->cell_adr_table[cur_cell].vob_id,
				vts_file->menu_c_adt->cell_adr_table[cur_cell].cell_id);
		printf("Sectors: 0x%08X to 0x%08X\n",
				vts_file->menu_c_adt->cell_adr_table[cur_cell].start_sector,
				vts_file->menu_c_adt->cell_adr_table[cur_cell].last_sector);
	}
	*/
    printf( "\n" );

    for( cur_cell = 0; cur_cell < entries; cur_cell ++ ) {
        /**
         * We loop until we're find a NAV pack.
         */
		buttons_found = 0;
        for( cur_pack = vts_file->menu_c_adt->cell_adr_table[cur_cell].start_sector;
             cur_pack <= vts_file->menu_c_adt->cell_adr_table[cur_cell].last_sector; ) {

            dsi_t dsi_pack;

	        /**
	         * Read NAV packet.
	         */
	        len = DVDReadBlocks( vts, (int) cur_pack, 1, data );
	        if ( len != 1 ) {
		  		printf( "%s%d,errorcode) 6\n", head, vtsnum );
		  		printf( "%s%d,errormsg) {Read failed for block %d. You need an unencrypted DVD.}\n",
		  				head, vtsnum, cur_pack);
	            ifoClose( vts_file );
	            DVDCloseFile( vts );
	            DVDClose( dvd );
	            return 0;
	        }
	        if ( is_nav_pack( data ) ) {
				if ( last_vob_id != vts_file->menu_c_adt->cell_adr_table[cur_cell].vob_id) {
			  		printf( "%s%d,%d,numcellids) %d\n\n", head, vtsnum, last_vob_id, last_cell_id );
					last_vob_id = vts_file->menu_c_adt->cell_adr_table[cur_cell].vob_id;
				}
				last_cell_id = vts_file->menu_c_adt->cell_adr_table[cur_cell].cell_id;

		        /**
		         * Parse the contained pci packet.
		         */
		        navRead_PCI( &pci_pack, &(data[ 0x2D ]) );
		        /* assert( cur_pack == pci_pack.dsi_gi.nv_pck_lbn ); */
		        if ( pci_pack.hli.hl_gi.btn_ns > 0 ) {
					printf( "\n# VOB/Cell ID: %d/%d\n", last_vob_id, last_cell_id );
			  		printf( "%s%d,%d,%d,navpack) %d\n", head, vtsnum, last_vob_id, last_cell_id, cur_pack );
		        	tcl_navPrint_BTNIT( &pci_pack.hli.hl_gi, pci_pack.hli.btnit, last_vob_id, last_cell_id, vtsnum);
		        	buttons_found = 1;
		        	break;
		        }
			}

            /**
             * Parse the contained dsi packet.
             */
            navRead_DSI( &dsi_pack, &(data[ DSI_START_BYTE ]) );
            cur_pack += dsi_pack.dsi_gi.vobu_ea + 1;
			/* printf("# next_pack: %d\n", cur_pack); */
		}
		if ( buttons_found == 0 ) {
			printf( "\n# VOB/Cell ID: %d/%d\n", last_vob_id, last_cell_id );
	  		printf( "%s%d,%d,%d,navpack) %d\n", head, vtsnum, last_vob_id, last_cell_id, \
	  				vts_file->menu_c_adt->cell_adr_table[cur_cell].start_sector );
        	tcl_navPrint_BTNIT( &pci_pack.hli.hl_gi, pci_pack.hli.btnit, last_vob_id, last_cell_id, vtsnum);
		}
    }
	printf( "%s%d,%d,numcellids) %d\n\n\n", head, vtsnum, last_vob_id, last_cell_id );
	printf( "%s%d,numvobids) %d\n\n", head, vtsnum, last_vob_id );

	printf( "%s%d,errorcode) 0\n", head, vtsnum );
	printf( "%s%d,errormsg) {OK}\n", head, vtsnum );
    ifoClose( vts_file );
    DVDCloseFile( vts );
    DVDClose( dvd );
    return 0;
}

